from __future__ import annotations

from openlifu.io.LIFUInterface import LIFUInterface

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/run_self_test.py
"""
Test script to automate:
1. Connect to the device.
2. Test HVController: Turn HV on/off and check voltage.
3. Test Device functionality.
"""
print("Starting LIFU Test Script...")
interface = LIFUInterface()
tx_connected, hv_connected = interface.is_device_connected()
if tx_connected and hv_connected:
    print("LIFU Device Fully connected.")
else:
    print(f'LIFU Device NOT Fully Connected. TX: {tx_connected}, HV: {hv_connected}')

print("Ping the device")
interface.txdevice.ping()

print("Run Self OneWire Test")
interface.txdevice.run_test()

print("Tests Finished")



from __future__ import annotations

from openlifu.io.LIFUInterface import LIFUInterface

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/test_updated_if.py
"""
Test script to automate:
1. Connect to the device.
2. Test HVController: Turn HV on/off and check voltage.
3. Test Device functionality.
"""
print("Starting LIFU Test Script...")
interface = LIFUInterface()
tx_connected, hv_connected = interface.is_device_connected()
if tx_connected and hv_connected:
    print("LIFU Device Fully connected.")
else:
    print(f'LIFU Device NOT Fully Connected. TX: {tx_connected}, HV: {hv_connected}')

print("Ping the device")
interface.txdevice.ping()

print("Get Temperature")
temperature = interface.txdevice.get_temperature()
print(f"Temperature: {temperature} °C")

print("Enumerate TX7332 chips")
num_tx_devices = interface.txdevice.enum_tx7332_devices()
if num_tx_devices > 0:
    print(f"Number of TX7332 devices found: {num_tx_devices}")
else:
    raise Exception("No TX7332 devices found.")

print("Set TX7332 Demo Waveform")
if interface.txdevice.demo_tx7332():
    print("TX7332 demo waveform set successfully.")
else:
    print("Failed to set TX7332 demo waveform.")

print("Get Trigger")
trigger_setting = interface.txdevice.get_trigger_json()
if trigger_setting:
    print(f"Trigger Setting: {trigger_setting}")
else:
    print("Failed to get trigger setting.")

print("Set Trigger")
json_trigger_data = {
    "TriggerFrequencyHz": 10,
    "TriggerMode": 1,
    "TriggerPulseCount": 0,
    "TriggerPulseWidthUsec": 20000
}
trigger_setting = interface.txdevice.set_trigger_json(data=json_trigger_data)
if trigger_setting:
    print(f"Trigger Setting: {trigger_setting}")
else:
    print("Failed to set trigger setting.")

print("Press enter to START trigger:")
input()  # Wait for the user to press Enter
print("Starting Trigger...")
if interface.txdevice.start_trigger():
    print("Trigger Running Press enter to STOP:")
    input()  # Wait for the user to press Enter
    if interface.txdevice.stop_trigger():
        print("Trigger stopped successfully.")
    else:
        print("Failed to stop trigger.")
else:
    print("Failed to get trigger setting.")



from __future__ import annotations

from dataclasses import asdict

import pytest

from openlifu.seg.material import Material

# Mock PARAM_INFO for tests
PARAM_INFO = {
    "name": {"label": "Material name", "description": "Name for the material"},
    "sound_speed": {"label": "Sound speed (m/s)", "description": "Speed of sound in the material (m/s)"},
    "density": {"label": "Density (kg/m^3)", "description": "Mass density of the material (kg/m^3)"},
    "attenuation": {"label": "Attenuation (dB/cm/MHz)", "description": "Ultrasound attenuation in the material (dB/cm/MHz)"},
    "specific_heat": {"label": "Specific heat (J/kg/K)", "description": "Specific heat capacity of the material (J/kg/K)"},
    "thermal_conductivity": {"label": "Thermal conductivity (W/m/K)", "description": "Thermal conductivity of the material (W/m/K)"},
}

@pytest.fixture()
def default_material():
    return Material()

def test_default_material_values(default_material):
    assert default_material.name == "Material"
    assert default_material.sound_speed == 1500.0
    assert default_material.density == 1000.0
    assert default_material.attenuation == 0.0
    assert default_material.specific_heat == 4182.0
    assert default_material.thermal_conductivity == 0.598

def test_material_to_dict(default_material):
    expected = asdict(default_material)
    assert default_material.to_dict() == expected

def test_material_from_dict():
    data = {
        "name": "test",
        "sound_speed": 1234.0,
        "density": 999.0,
        "attenuation": 1.2,
        "specific_heat": 4000.0,
        "thermal_conductivity": 0.5
    }
    material = Material.from_dict(data)
    assert material.to_dict() == data

def test_get_param_valid(default_material):
    assert default_material.get_param("density") == 1000.0

def test_get_param_invalid(default_material):
    with pytest.raises(ValueError, match="Parameter fake_param not found."):
        default_material.get_param("fake_param")

def test_param_info_valid(monkeypatch):
    monkeypatch.setattr("openlifu.seg.material.PARAM_INFO", PARAM_INFO)
    info = Material.param_info("density")
    assert info["label"] == "Density (kg/m^3)"

def test_param_info_invalid(monkeypatch):
    monkeypatch.setattr("openlifu.seg.material.PARAM_INFO", PARAM_INFO)
    with pytest.raises(ValueError, match="Parameter unknown not found."):
        Material.param_info("unknown")



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# Once the environment has been set up, this notebook should be able to be run, cell-by-cell. One issue is that I've made modifications to k-wave-python (to cache and reuse gridweights, which saves a significant amount of time if you are running different simulations using the same array in the same position - it's being discussed in https://github.com/waltsims/k-wave-python/issues/342, but hasn't been merged yet, afaik).
#
# If you have my modified version of k-wave-python, install it with `pip install -e .` from the `k-wave-python` directory. If you don't have my modified version, you can install the original version with `pip install k-wave-python`. If you are using the original version, be sure to set `USE_GRIDWEIGHTS` to `False` in order to prevent `open_pyfus` from trying to use a nonexistent interface for loading the gridweights.
#
# Also, if you are using the original version, import `openlifu` takes _way_ longer (45s on my PC), presumably hanging on `import kwave`. For some reason, it wants to re-download the binaries every time, even though they are already present in the the installation directory. I've opened an issue on this: https://github.com/waltsims/k-wave-python/issues/366.

# %%
from __future__ import annotations

import logging
import sys

root = logging.getLogger()
loglevel = logging.DEBUG
root.setLevel(loglevel)
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(loglevel)
formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
handler.setFormatter(formatter)
root.addHandler(handler)
import numpy as np

import openlifu

# %% [markdown]
# We'll start by generating a transducer and drawing it using some vtk-based methods

# %%
arr = openlifu.Transducer.gen_matrix_array(nx=8, ny=8, pitch=4, kerf=.5, units="mm", impulse_response=1e5)
arr.draw()

# %% [markdown]
# Now we'll define the pulse and sequence parameters, the simulation setup, and generate a Protocol.

# %%
pulse = openlifu.Pulse(frequency=400e3, duration=3/400e3)
sequence = openlifu.Sequence()
focal_pattern = openlifu.focal_patterns.Wheel(center=True, spoke_radius=5, num_spokes=5)
sim_setup = openlifu.SimSetup(dt=2e-7, t_end=100e-6)
protocol = openlifu.Protocol(
    pulse=pulse,
    sequence=sequence,
    focal_pattern=focal_pattern,
    sim_setup=sim_setup)

# %% [markdown]
# Now we can define a sonication target(s), set up the simulation, and compute the delays and apodizations needed to steer the sound to it.

# %%
pt = openlifu.Point(position=(0,0,30), units="mm", radius=2)
pts = protocol.focal_pattern.get_targets(pt)
coords = protocol.sim_setup.get_coords()
params = protocol.seg_method.ref_params(coords)
delays, apod = protocol.beamform(arr=arr, target=pts[0], params=params)

# %% [markdown]
# Now we are ready to run the simulation.  Some custom edits to `k-wave-python` allow for caching of the gridweights, which only need to be computed once for a given grid size and source location.  This can speed up the simulation significantly, especially if a coarse grid that won't take the GPU too long to run is used.

# %%
(ds, output) = openlifu.sim.run_simulation(arr=arr,
        params=params,
        delays=delays,
        apod= apod,
        freq = pulse.frequency,
        cycles = np.max([np.round(pulse.duration * pulse.frequency), 20]),
        dt=protocol.sim_setup.dt,
        t_end=protocol.sim_setup.t_end,
        amplitude = 1,
        gpu = False)

# %% [markdown]
# We can use all of `xarray`s built-in plotting capabilities to plot the data.

# %%
ds['p_min'].sel(ele=0).plot.imshow()

# %%
ds['p_min'].sel(ele=0).plot.imshow()

# %% [markdown]
# We can examine the output object, which is an `xarray.DataSet` object with 3 data variables: `p_max` (Peak Positive Pressure), `p_min` (Peak Negative Pressure), and `intensity` (Time Averaged Intensity). It's attributes also contain the `source` pulse (an `xarray.DataArray`), and `output`, the raw K-Wave output structure.

# %%
ds

# %% [markdown]
# Using `nibabel`, we can export the DataArray to a NIftI file. This requires a little bit of manipulation of the coordinates to extract the origin and affine matrix as NIftI needs them. This should get wrapped into a function in the future.

# %%
import nibabel as nb

p_min = ds['p_min'].data
coords = ds['p_min'].coords
affine = np.eye(3) * np.array([float(np.diff(coords[x][:2])) for x in coords])
origin = np.array([float(coords[x][0]) for x in coords]).reshape(3,1)
affine = np.concatenate([np.concatenate([affine, origin], axis=1),np.array([0,0,0,1]).reshape(1,4)], axis=0)
nb.Nifti1Image(p_min, affine).to_filename("p_min.nii.gz")

# %% [markdown]
# Finally, we can use some of the intermediate vtk methods to extract Actors from both the array and points objects, and pipe them to a since render:

# %%
import vtk

arr_actor = arr.get_actor(units="mm")
renderWindow = vtk.vtkRenderWindow()
renderer = vtk.vtkRenderer()
renderWindow.AddRenderer(renderer)
renderWindowInteractor = vtk.vtkRenderWindowInteractor()
renderWindowInteractor.SetRenderWindow(renderWindow)
renderer.AddActor(arr_actor)
for pti in pts:
    pt_actor = pti.get_actor()
    renderer.AddActor(pt_actor)
renderWindow.Render()
renderWindowInteractor.Start()



from __future__ import annotations

import numpy as np
import pytest
from xarray import DataArray, Dataset

from openlifu.util.units import (
    get_ndgrid_from_arr,
    getsiscale,
    rescale_coords,
    rescale_data_arr,
)


@pytest.fixture()
def example_xarr() -> Dataset:
    rng = np.random.default_rng(147)
    return Dataset(
            {
                'p': DataArray(
                    data=rng.random((3, 2)),
                    dims=["x", "y"],
                    attrs={'units': "Pa"}
                ),
                'it': DataArray(
                    data=rng.random((3, 2)),
                    dims=["x", "y"],
                    attrs={'units': "W/cm^2"}
                )
            },
            coords={
                'x': DataArray(dims=["x"], data=np.linspace(0, 1, 3), attrs={'units': "m"}),
                'y': DataArray(dims=["y"], data=np.linspace(0, 1, 2), attrs={'units': "m"})
            }
        )


def test_getsiscale():
    with pytest.raises(ValueError, match="Unknown prefix"):
        getsiscale('xx','distance')

    assert getsiscale('mm', 'distance') == 1e-3
    assert getsiscale('km', 'distance') == 1e3
    assert getsiscale('mm^2', 'area') == 1e-6
    assert getsiscale('mm^3', 'volume') == 1e-9
    assert getsiscale('ns', 'time') == 1e-9
    assert getsiscale('nanosecond', 'time') == 1e-9
    assert getsiscale('hour', 'time') == 3600.
    assert getsiscale('rad', 'angle') == 1.
    assert np.allclose(getsiscale('deg', 'angle'), np.pi/180.)
    assert getsiscale('MHz', 'frequency') == 1e6
    assert getsiscale('GHz', 'frequency') == 1e9
    assert getsiscale('THz', 'frequency') == 1e12


def test_rescale_data_arr(example_xarr: Dataset):
    """Test that an xarray data can be correctly rescaled."""
    expected_p = 1e-6 * example_xarr['p'].data
    expected_it = 1e4 * example_xarr['it'].data
    rescaled_p = rescale_data_arr(example_xarr['p'], units="MPa")
    rescaled_it = rescale_data_arr(example_xarr['it'], units="W/m^2")

    np.testing.assert_almost_equal(rescaled_p, expected_p)
    np.testing.assert_almost_equal(rescaled_it, expected_it)


def test_rescale_coords(example_xarr: Dataset):
    """Test that an xarray coords can be correctly rescaled."""
    expected_x = 1e3 * example_xarr['p'].coords['x'].data
    expected_y = 1e3 * example_xarr['p'].coords['y'].data
    rescaled = rescale_coords(example_xarr['p'], units="mm")

    np.testing.assert_almost_equal(rescaled['x'].data, expected_x)
    np.testing.assert_almost_equal(rescaled['y'].data, expected_y)


def test_get_ndgrid_from_arr(example_xarr: Dataset):
    """Test that an ndgrid can be constructed from the coordinates of an xarray."""
    expected_X = np.array([[0., 0.], [0.5, 0.5], [1., 1.]])
    expected_Y = np.array([[0., 1.], [0., 1.], [0., 1.]])
    expected = np.stack([expected_X, expected_Y], axis=-1)
    ndgrid = get_ndgrid_from_arr(example_xarr)

    np.testing.assert_equal(ndgrid, expected)



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Annotated

from openlifu.bf import focal_patterns
from openlifu.geo import Point
from openlifu.util.annotations import OpenLIFUFieldData


@dataclass
class FocalPattern(ABC):
    """
    Abstract base class for representing a focal pattern
    """

    target_pressure: Annotated[float, OpenLIFUFieldData("Target pressure", "Target pressure of the focal pattern in given units")] = 1.0
    """Target pressure of the focal pattern in given units"""

    units: Annotated[str, OpenLIFUFieldData("Pressure units", "Pressure units")] = "Pa"
    """Pressure units"""

    @abstractmethod
    def get_targets(self, target: Point):
        """
        Get the targets of the focal pattern

        :param target: Target point of the focal pattern
        :returns: List of target points
        """
        pass

    @abstractmethod
    def num_foci(self):
        """
        Get the number of foci in the focal pattern

        :returns: Number of foci
        """
        pass

    def to_dict(self):
        """
        Convert the focal pattern to a dictionary

        :returns: Dictionary of the focal pattern parameters
        """
        d = self.__dict__.copy()
        d['class'] = self.__class__.__name__
        return d

    @staticmethod
    def from_dict(d):
        """
        Create a focal pattern from a dictionary

        :param d: Dictionary of the focal pattern parameters
        :returns: FocalPattern object
        """
        d = d.copy()
        short_classname = d.pop("class")
        module_dict = focal_patterns.__dict__
        class_constructor = module_dict[short_classname]
        return class_constructor(**d)



from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated

import numpy as np
import xarray as xa

from openlifu.bf.delay_methods import DelayMethod
from openlifu.geo import Point
from openlifu.util.annotations import OpenLIFUFieldData
from openlifu.xdc import Transducer


@dataclass
class Direct(DelayMethod):
    c0: Annotated[float, OpenLIFUFieldData("Speed of Sound (m/s)", "Speed of sound in the medium (m/s)")] = 1480.0
    """Speed of sound in the medium (m/s)"""

    def calc_delays(self, arr: Transducer, target: Point, params: xa.Dataset | None=None, transform:np.ndarray | None=None):
        if params is None:
            c = self.c0
        else:
            c = params['sound_speed'].attrs['ref_value']
        target_pos = target.get_position(units="m")
        matrix = transform if transform is not None else np.eye(4)
        dists = np.array([el.distance_to_point(target_pos, units="m", matrix=matrix) for el in arr.elements])
        tof = dists / c
        delays = max(tof) - tof
        return delays



from __future__ import annotations

import numpy as np

from openlifu.geo import (
    Point,
    cartesian_to_spherical,
    cartesian_to_spherical_vectorized,
    create_standoff_transform,
    spherical_coordinate_basis,
    spherical_to_cartesian,
    spherical_to_cartesian_vectorized,
)


def test_point_from_dict():
    point = Point.from_dict({'position' : [10,20,30],})
    assert (point.position == np.array([10,20,30], dtype=float)).all()

def test_spherical_coordinate_range():
    """Verify that spherical coordinate output is in the prescribed value ranges"""
    rng = np.random.default_rng(848)
    # try all 8 octants of 3D space
    for sign_x in [-1,1]:
        for sign_y in [-1,1]:
            for sign_z in [-1,1]:
                cartesian_coords = np.array([sign_x, sign_y, sign_z]) * rng.random(size=3)
                r, th, ph = cartesian_to_spherical(*cartesian_coords)
                assert r>=0
                assert 0 <= th <= np.pi
                assert -np.pi <= ph <= np.pi

def test_spherical_coordinate_conversion_inverse():
    """Verify that the spherical coordinate conversion forward and backward functions are inverses of one another"""
    rng = np.random.default_rng(241)
    # try all 8 octants of 3D space
    for sign_x in [-1,1]:
        for sign_y in [-1,1]:
            for sign_z in [-1,1]:
                cartesian_coords = np.array([sign_x, sign_y, sign_z]) * rng.random(size=3)
                np.testing.assert_almost_equal(
                    spherical_to_cartesian(*cartesian_to_spherical(*cartesian_coords)),
                    cartesian_coords
                )
                np.testing.assert_almost_equal(
                    cartesian_to_spherical(*spherical_to_cartesian(*cartesian_to_spherical(*cartesian_coords))),
                    cartesian_to_spherical(*cartesian_coords)
                )

def test_cartesian_to_spherical_vectorized():
    rng = np.random.default_rng(35932)
    points_cartesian = rng.normal(size=(10,3), scale=2) # make 10 random cartesian points
    points_spherical = cartesian_to_spherical_vectorized(points_cartesian)
    # Check individual points against the non-vectorized conversion function:
    for point_cartesian, point_spherical in zip(points_cartesian, points_spherical):
        assert np.allclose(
            point_spherical, # result of vectorized converter
            np.array(cartesian_to_spherical(*point_cartesian)), # non-vectorized converter
        )

def test_spherical_to_cartesian_vectorized():
    rng = np.random.default_rng(85932)

    # make 10 random points in spherical coordinates
    num_pts = 10
    points_spherical = np.zeros(shape=(num_pts,3))
    points_spherical[...,0] = rng.random(num_pts)*5 # random r coordinates
    points_spherical[...,1] = rng.random(num_pts)*np.pi # random theta coordinates
    points_spherical[...,2] = rng.random(num_pts)*2*np.pi-np.pi # random phi coordinates

    points_cartesian = spherical_to_cartesian_vectorized(points_spherical)
    # Check individual points against the non-vectorized conversion function:
    for point_cartesian, point_spherical in zip(points_cartesian, points_spherical):
        assert np.allclose(
            point_cartesian, # result of vectorized converter
            np.array(spherical_to_cartesian(*point_spherical)), # non-vectorized converter
        )

def test_spherical_coordinate_basis():
    rng = np.random.default_rng(35235)
    th = rng.random()*np.pi
    phi = rng.random()*2*np.pi-np.pi
    r  = rng.random()*10
    basis = spherical_coordinate_basis(th,phi)
    assert np.allclose(basis @ basis.T, np.eye(3)) # verify it is an orthonormal basis
    r_hat, theta_hat, phi_hat = basis
    point = np.array(spherical_to_cartesian(r, th, phi))
    assert np.allclose(np.diff(r_hat / point), 0) # verify that r_hat is a scalar multiple of the cartesian coords
    assert cartesian_to_spherical_vectorized(point + 0.01*phi_hat)[2] > phi # verify phi_hat points along increasing phi
    assert cartesian_to_spherical_vectorized(point + 0.01*theta_hat)[1] > th # verify theta_hat points along increasing theta

def test_create_standoff_transform():
    z_offset = 3.2
    dzdy = 0.15
    t = create_standoff_transform(z_offset, dzdy)
    assert np.allclose(t[:3,:3] @ t[:3,:3].T, np.eye(3)) # it's an orthonormal transform
    assert np.allclose(np.linalg.det(t[:3,:3]), 1.0) # orientation preserving
    assert np.allclose(t @ np.array([0,0,0,1]), np.array([0,0,-z_offset,1.])) # translates the origin correctly
    new_x_axis = (t @ np.array([1,0,0,1]) - t @ np.array([0,0,0,1]))[:3]
    new_y_axis = (t @ np.array([0,1,0,1]) - t @ np.array([0,0,0,1]))[:3]
    assert np.allclose(new_x_axis, np.array([1.,0,0]))
    assert new_y_axis[2] > 0 # the y axis was rotated upward, so that the top of the transducer gets closer to the skin



from __future__ import annotations

from dataclasses import asdict, dataclass, fields
from typing import Any, Dict, Type, TypeVar, get_origin

import numpy as np

T = TypeVar('T', bound='DictMixin')

@dataclass
class DictMixin:
    """Mixin for basic conversion of a dataclass to and from dict."""
    def to_dict(self) -> Dict[str,Any]:
        """
        Convert the object to a dictionary

        Returns: Dictionary of object parameters
        """
        return asdict(self)

    @classmethod
    def from_dict(cls : Type[T], parameter_dict:Dict[str,Any]) -> T:
        """
        Create an object from a dictionary

        Args:
            parameter_dict: dictionary of parameters to define the object
        Returns: new object
        """
        if "class" in parameter_dict:
            parameter_dict.pop("class")
        new_object = cls(**parameter_dict)

        # Convert anything that should be a numpy array to numpy
        for field in fields(cls):
            # Note that sometimes "field.type" is a string rather than a type due to the "from annotations import __future__" stuff
            if get_origin(field.type) is np.ndarray or field.type is np.ndarray or "np.ndarray" in field.type:
                setattr(new_object, field.name, np.array(getattr(new_object,field.name)))

        return new_object



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: env
#     language: python
#     name: python3
# ---

# %%
from __future__ import annotations

import numpy as np

from openlifu.io.LIFUTXDevice import (
    Tx7332DelayProfile,
    Tx7332PulseProfile,
    Tx7332Registers,
    TxDeviceRegisters,
    print_regs,
)

# %%
tx = Tx7332Registers()
delays = np.arange(32)*1e-6
apodizations = np.ones(32)
delay_profile_1 = Tx7332DelayProfile(1, delays, apodizations)
tx.add_delay_profile(delay_profile_1)
print('CONTROL')
print_regs(tx.get_delay_control_registers())
print('DATA')
print_regs(tx.get_delay_data_registers())

# %%
frequency = 150e3
cycles = 3
pulse_profile_1 = Tx7332PulseProfile(1, frequency, cycles)
print('CONTROL')
tx.add_pulse_profile(pulse_profile_1)
print_regs(tx.get_pulse_control_registers())
print('DATA')
print_regs(tx.get_pulse_data_registers())

# %%
tx.add_delay_profile(delay_profile_1)
tx.add_pulse_profile(pulse_profile_1)
r = tx.get_registers(profiles="all")
print_regs(r)

# %%
x = np.linspace(-0.5, 0.5, 32)*4e-2
r = np.sqrt(x**2 + 5e-2**2)
delays = (r.max()-r)/1500
apodizations = [0,1]*16
delay_profile_2 = Tx7332DelayProfile(2, delays, apodizations)
tx.add_delay_profile(delay_profile_2)
print(f'{len(tx._delay_profiles_list)} Delay Profiles')
print(f'{len(tx._pulse_profiles_list)} Pulse Profiles')
print('')
for pack in [False, True]:
    print(f'Pack: {pack}')
    for profile_opts in ["active", "configured", "all"]:
        r = tx.get_registers(profiles=profile_opts, pack=pack)
        print(f'{profile_opts}: {len(r)} Writes')
    print('')

# %%
for index in [1,2]:
    tx.activate_delay_profile(index)
    print('\n')
    print(index)
    print('CONTROL')
    print_regs(tx.get_delay_control_registers())
    print('DATA')
    print_regs(tx.get_delay_data_registers())


# %%
txm = TxDeviceRegisters()
delays = np.arange(64)*1e-6
apodizations = np.ones(64)
module_delay_profile_1 = Tx7332DelayProfile(1, delays, apodizations)
txm.add_delay_profile(module_delay_profile_1)
txm.add_pulse_profile(pulse_profile_1)
for i, r in enumerate(txm.get_delay_control_registers()):
    print(f'CONTROL {i}')
    print_regs(r)
for i, r in enumerate(txm.get_delay_data_registers()):
    print(f'DATA {i}')
    print_regs(r)

# %%
module_delay_profile_1

# %%
x = np.linspace(-0.5, 0.5, 64)*4e-2
r = np.sqrt(x**2 + 5e-2**2)
delays = (r.max()-r)/1500
apodizations = [0,1]*32
module_delay_profile_2 = Tx7332DelayProfile(2, delays, apodizations)
frequency = 100e3
cycles = 200
pulse_profile_2 = Tx7332PulseProfile(2, frequency, cycles)
txm.add_delay_profile(module_delay_profile_2, activate=True)
txm.add_pulse_profile(pulse_profile_2, activate=True)
for i, r in enumerate(txm.get_delay_control_registers()):
    print(f'DELAY CONTROL {i}')
    print_regs(r)
for i, r in enumerate(txm.get_delay_data_registers()):
    print(f'DELAY DATA {i}')
    print_regs(r)

# %%
for i, r in enumerate(txm.get_pulse_control_registers()):
    print(f'PULSE CONTROL {i}')
    print_regs(r)
for i, r in enumerate(txm.get_pulse_data_registers()):
    print(f'PULSE DATA {i}')
    print_regs(r)

# %%
r = {'DELAY CONTROL': {}, 'DELAY DATA': {}, 'PULSE CONTROL': {}, 'PULSE DATA': {}}
rtemplate = {}
profiles = [1,2]
for index in profiles:
    rtemplate[index] = ['---------']*2
for index in profiles:
    txm.activate_delay_profile(index)
    txm.activate_pulse_profile(index)
    rdcm = txm.get_delay_control_registers()
    rddm = txm.get_delay_data_registers()
    rpcm = txm.get_pulse_control_registers()
    rpdm = txm.get_pulse_data_registers()
    d = {'DELAY CONTROL': rdcm, 'DELAY DATA': rddm, 'PULSE CONTROL': rpcm, 'PULSE DATA': rpdm}
    for k, rm in d.items():
        for txi, rc in enumerate(rm):
            for addr, value in rc.items():
                if addr not in r[k]:
                    r[k][addr] = {i: ['---------']*2 for i in profiles}
                r[k][addr][index][txi] = f'x{value:08x}'
h = [f'{"Profile " + str(i):19s}' for i in profiles]
print(f"      {' | '.join(h)}")
h1 = [f'{"TX" + str(i):>9s}' for i in [0,1]]
h1s = [' '.join(h1)]*len(profiles)
print(f"addr: {' | '.join(h1s)}")
for k, rm in r.items():
    print(f"{k}")
    for addr, rr in rm.items():
        print(f"x{addr:03x}: {' | '.join([' '.join(rr[i]) for i in profiles])}")

# %%
txm.get_registers(pack=False, pack_single=False)



from __future__ import annotations

import copy
from collections.abc import Iterable
from dataclasses import dataclass, field
from typing import Annotated

import numpy as np

from openlifu.util.annotations import OpenLIFUFieldData
from openlifu.util.units import getunitconversion


def matrix2xyz(matrix):
    x = matrix[0, 3]
    y = matrix[1, 3]
    z = matrix[2, 3]
    az = np.arctan2(matrix[0, 2], matrix[2, 2])
    el = -np.arctan2(matrix[1, 2], np.sqrt(matrix[2, 2]**2 + matrix[0, 2]**2))
    Raz = np.array([[np.cos(az), 0, np.sin(az)],
                    [0, 1, 0],
                    [-np.sin(az), 0, np.cos(az)]])
    Rel = np.array([[1, 0, 0],
                    [0, np.cos(el), -np.sin(el)],
                    [0, np.sin(el), np.cos(el)]])
    Razel = np.dot(Raz, Rel)
    xv = matrix[:3, 0]
    xyp = np.dot(xv, Razel[:3,1])
    xxp = np.dot(xv, Razel[:3,0])
    roll = np.arctan2(xyp, xxp)
    return x, y, z, az, el, roll

@dataclass
class Element:
    index: Annotated[int, OpenLIFUFieldData("Element index", "Element index")] = 0
    """Element index to identify the element in the array."""

    x: Annotated[float, OpenLIFUFieldData("X position", "X position of the element")] = 0
    """X position of the element."""

    y: Annotated[float, OpenLIFUFieldData("Y position", "Y position of the element")] = 0
    """Y position of the element."""

    z: Annotated[float, OpenLIFUFieldData("Z position", "Z position of the element")] = 0
    """Z position of the element."""

    az: Annotated[float, OpenLIFUFieldData("Azimuth angle (rad)", "Azimuth angle of the element")] = 0
    """Azimuth angle of the element, or rotation about the y-axis (rad)"""

    el: Annotated[float, OpenLIFUFieldData("Elevation angle (rad)", "Elevation angle of the element")] = 0
    """Elevation angle of the element, or rotation about the x'-axis (rad)"""

    roll: Annotated[float, OpenLIFUFieldData("Roll angle (rad)", "Roll angle of the element")] = 0
    """Roll angle of the element, or rotation about the z''-axis (rad)"""

    w: Annotated[float, OpenLIFUFieldData("Width", "Width of the element in the x dimension")] = 1
    """Width of the element in the x dimension"""

    l: Annotated[float, OpenLIFUFieldData("Length", "Length of the element in the y dimension")] = 1
    """Length of the element in the y dimension"""

    impulse_response: Annotated[np.ndarray, OpenLIFUFieldData("Impulse response", "Impulse response of the element")] = field(repr=False, default_factory=lambda: np.array([1]))
    """Impulse response of the element, can be a single value or an array of values. If an array, `impulse_dt` must be set to the time step of the impulse response. Is convolved with the input signal."""

    impulse_dt: Annotated[float, OpenLIFUFieldData("Impulse response timestep", """Impulse response timestep""")] = field(repr=False, default=1)
    """Impulse response timestep. If `impulse_response` is an array, this is the time step of the impulse response."""

    pin: Annotated[int, OpenLIFUFieldData("Pin", "Channel pin to which the element is connected")] = -1
    """Channel pin to which the element is connected. 1-(64*number of modules)."""

    units: Annotated[str, OpenLIFUFieldData("Units", "Spatial units")] = "mm"
    """Spatial units of the element specification."""

    def __post_init__(self):
        if isinstance(self.impulse_response, Iterable):
            self.impulse_response = np.array(self.impulse_response, dtype=np.float64)
        else:
            self.impulse_response = np.array([self.impulse_response], dtype=np.float64)

    def calc_output(self, input_signal, dt):
        if len(self.impulse_response) == 1:
            return input_signal * self.impulse_response
        else:
            impulse = self.interp_impulse_response(dt)
            return np.convolve(input_signal, impulse, mode='full')

    def copy(self):
        return copy.deepcopy(self)

    def rescale(self, units):
        if self.units != units:
            scl = getunitconversion(self.units, units)
            self.x *= scl
            self.y *= scl
            self.z *= scl
            self.w *= scl
            self.l *= scl
            self.units = units

    def get_position(self, units=None, matrix=np.eye(4)):
        units = self.units if units is None else units
        scl = getunitconversion(self.units, units)
        pos = np.array([self.x, self.y, self.z]) * scl
        pos = np.append(pos, 1)
        pos = np.dot(matrix, pos)
        return pos[:3]

    def get_size(self, units=None):
        units = self.units if units is None else units
        scl = getunitconversion(self.units, units)
        ele_width = self.w * scl
        ele_length = self.l * scl
        return ele_width, ele_length

    def get_area(self, units=None):
        units = self.units if units is None else units
        ele_width, ele_length = self.get_size(units)
        return ele_width * ele_length

    def get_corners(self, units=None, matrix=np.eye(4)):
        units = self.units if units is None else units
        scl = getunitconversion(self.units, units)
        rect = np.array([np.array([-1, -1.,  1,  1]) * 0.5 * self.w,
                            np.array([-1,  1,  1, -1]) * 0.5 * self.l,
                            np.zeros(4) ,
                            np.ones(4)])
        xyz = np.dot(self.get_matrix(), rect)
        xyz1 = np.dot(matrix, xyz)
        corner = []
        for j in range(3):
            corner.append(xyz1[j, :] * scl)
        return np.array(corner            )

    def get_matrix(self, units=None):
        units = self.units if units is None else units
        Raz = np.array([[np.cos(self.az), 0, np.sin(self.az)],
                        [0, 1, 0],
                        [-np.sin(self.az), 0, np.cos(self.az)]])
        Rel = np.array([[1, 0, 0],
                        [0, np.cos(self.el), -np.sin(self.el)],
                        [0, np.sin(self.el), np.cos(self.el)]])
        Rroll = np.array([[np.cos(self.roll), -np.sin(self.roll), 0],
                            [np.sin(self.roll), np.cos(self.roll), 0],
                            [0, 0, 1]])
        pos = self.get_position(units=units)
        m = np.concatenate((np.dot(Raz, np.dot(Rel,Rroll)), pos.reshape([3,1])), axis=1)
        m = np.concatenate((m, [[0, 0, 0, 1]]), axis=0)
        return m

    def get_angle(self, units="rad"):
        # Return angles about the x, y', and z'' axes (el, az, roll)
        if units == "rad":
            el = self.el
            az = self.az
            roll = self.roll
        elif units == "deg":
            el = np.degrees(self.el)
            az = np.degrees(self.az)
            roll = np.degrees(self.roll)
        return el, az, roll

    def interp_impulse_response(self, dt=None):
        if dt is None:
            dt = self.impulse_dt
        n0 = len(self.impulse_response)
        if n0 == 1:
            impulse_response= self.impulse_response
        else:
            t0 = self.impulse_dt * np.arange(n0)
            t1 = np.arange(0, t0[-1] + dt, dt)
            impulse_response = np.interp(t1, t0, self.impulse_response)
        impulse_t = np.arange(len(impulse_response)) * dt
        impulse_t = impulse_t - np.mean(impulse_t)
        return impulse_response, impulse_t

    def distance_to_point(self, point, units=None, matrix=np.eye(4)):
        units = self.units if units is None else units
        pos = np.concatenate([self.get_position(units=units), [1]])
        m = self.get_matrix(units=units)
        gpos = np.dot(matrix, pos)
        vec = point - gpos[:3]
        dist = np.linalg.norm(vec, 2)
        return dist

    def angle_to_point(self, point, units=None, return_as="rad", matrix=np.eye(4)):
        units = self.units if units is None else units
        m = self.get_matrix(units=units)
        gm = np.dot(matrix, m)
        v1 = point - gm[:3, 3]
        v2 = gm[:3, 2]
        v1 = v1 / np.linalg.norm(v1, 2)
        v2 = v2 / np.linalg.norm(v2, 2)
        vcross = np.cross(v1, v2)
        theta = np.arcsin(np.linalg.norm(vcross, 2))
        if return_as == "deg":
            theta = np.degrees(theta)
        return theta

    def set_matrix(self, matrix, units=None):
        if units is not None:
            self.rescale(units)
        x, y, z, az, el, roll = matrix2xyz(matrix)
        self.x = x
        self.y = y
        self.z = z
        self.az = az
        self.el = el
        self.roll = roll

    def to_dict(self):
        return {"index": self.index,
                "x": self.x,
                "y": self.y,
                "z": self.z,
                "az": self.az,
                "el": self.el,
                "roll": self.roll,
                "w": self.w,
                "l": self.l,
                "impulse_response": self.impulse_response.tolist(),
                "impulse_dt": self.impulse_dt,
                "pin": self.pin,
                "units": self.units}

    @staticmethod
    def from_dict(d):
        return Element(**d)



from __future__ import annotations

import numpy as np
import pytest
from helpers import dataclasses_are_equal

from openlifu import Point


@pytest.fixture()
def example_point() -> Point:
    return Point(
        id = "example_point",
        name="Example point",
        color=(0.,0.7, 0.2),
        radius=1.5,
        position=np.array([-10.,0,25]),
        dims = ("R", "A", "S"),
        units = "m",
    )

@pytest.mark.parametrize("compact_representation", [True, False])
@pytest.mark.parametrize("default_point", [True, False])
def test_serialize_deserialize_point(example_point : Point, compact_representation: bool, default_point: bool):
    """Verify that turning a point into json and then re-constructing it gets back to the original point"""
    point = Point() if default_point else example_point
    reconstructed_point = point.from_json(point.to_json(compact_representation))
    assert dataclasses_are_equal(point, reconstructed_point)



from __future__ import annotations

import logging
import struct

from openlifu.io.LIFUConfig import (
    OW_CMD,
    OW_CMD_DFU,
    OW_CMD_ECHO,
    OW_CMD_HWID,
    OW_CMD_PING,
    OW_CMD_RESET,
    OW_CMD_TOGGLE_LED,
    OW_CMD_VERSION,
    OW_ERROR,
    OW_POWER,
    OW_POWER_12V_OFF,
    OW_POWER_12V_ON,
    OW_POWER_GET_12VON,
    OW_POWER_GET_FAN,
    OW_POWER_GET_HV,
    OW_POWER_GET_HVON,
    OW_POWER_GET_RGB,
    OW_POWER_GET_TEMP1,
    OW_POWER_GET_TEMP2,
    OW_POWER_HV_OFF,
    OW_POWER_HV_ON,
    OW_POWER_SET_DACS,
    OW_POWER_SET_FAN,
    OW_POWER_SET_HV,
    OW_POWER_SET_RGB,
)
from openlifu.io.LIFUUart import LIFUUart

logger = logging.getLogger(__name__)


class HVController:
    def __init__(self, uart: LIFUUart = None):
        """
        Initialize the HVController.

        Args:
            uart (LIFUUart): The LIFUUart instance for communication.
        """
        self.uart = uart

        if self.uart and not self.uart.asyncMode:
            self.uart.check_usb_status()
            if self.uart.is_connected():
                logger.info("HV Console connected.")
            else:
                logger.info("HV Console NOT Connected.")

        # Initialize the high voltage state (should get this from device)
        self.output_voltage = 0.0
        self.is_hv_on = False
        self.is_12v_on = False

        self.supply_voltage = None

    def is_connected(self):
        if self.uart:
            return self.uart.is_connected()

    def ping(self) -> bool:
        """
        Send a ping command to the Console device to verify connectivity.

        Raises:
            ValueError: If the UART is not connected.
            Exception: If an error occurs during the ping process.
        """
        try:
            if self.uart.demo_mode:
                return True

            if not self.uart.is_connected():
                raise ValueError("Console Device not connected")

            logger.info("Send Ping to Device.")
            r = self.uart.send_packet(id=None, packetType=OW_CMD, command=OW_CMD_PING)
            self.uart.clear_buffer()
            logger.info("Received Ping from Device.")
            # r.print_packet()

            if r.packet_type == OW_ERROR:
                logger.error("Error sending ping")
                return False
            else:
                return True

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def get_version(self) -> str:
        """
        Retrieve the firmware version of the Console device.

        Returns:
            str: Firmware version in the format 'vX.Y.Z'.

        Raises:
            ValueError: If the UART is not connected.
            Exception: If an error occurs while fetching the version.
        """
        try:
            if self.uart.demo_mode:
                return "v0.1.1"

            if not self.uart.is_connected():
                raise ValueError("Console Device not connected")

            r = self.uart.send_packet(
                id=None, packetType=OW_CMD, command=OW_CMD_VERSION
            )
            self.uart.clear_buffer()
            # r.print_packet()
            if r.data_len == 3:
                ver = f"v{r.data[0]}.{r.data[1]}.{r.data[2]}"
            else:
                ver = "v0.0.0"
            logger.info(ver)
            return ver

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def echo(self, echo_data=None) -> tuple[bytes, int]:
        """
        Send an echo command to the device with data and receive the same data in response.

        Args:
            echo_data (bytes): The data to send (must be a byte array).

        Returns:
            tuple[bytes, int]: The echoed data and its length.

        Raises:
            ValueError: If the UART is not connected.
            TypeError: If the `echo_data` is not a byte array.
            Exception: If an error occurs during the echo process.
        """
        try:
            if self.uart.demo_mode:
                data = b"Hello LIFU!"
                return data, len(data)

            if not self.uart.is_connected():
                raise ValueError("Console Device  not connected")

            # Check if echo_data is a byte array
            if echo_data is not None and not isinstance(echo_data, (bytes, bytearray)):
                raise TypeError("echo_data must be a byte array")

            r = self.uart.send_packet(
                id=None, packetType=OW_CMD, command=OW_CMD_ECHO, data=echo_data
            )
            self.uart.clear_buffer()
            # r.print_packet()
            if r.data_len > 0:
                return r.data, r.data_len
            else:
                return None, None

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def toggle_led(self) -> bool:
        """
        Toggle the LED on the Console device.

        Raises:
            ValueError: If the UART is not connected.
            Exception: If an error occurs while toggling the LED.
        """
        try:
            if self.uart.demo_mode:
                return True

            if not self.uart.is_connected():
                raise ValueError("Console Device not connected")

            r = self.uart.send_packet(
                id=None, packetType=OW_CMD, command=OW_CMD_TOGGLE_LED
            )
            self.uart.clear_buffer()
            # r.print_packet()
            if r.packet_type == OW_ERROR:
                return False

            return True
        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def get_hardware_id(self) -> str:
        """
        Retrieve the hardware ID of the Console device.

        Returns:
            str: Hardware ID in hexadecimal format.

        Raises:
            ValueError: If the UART is not connected.
            Exception: If an error occurs while retrieving the hardware ID.
        """
        try:
            if self.uart.demo_mode:
                return bytes.fromhex("deadbeefcafebabe5566778811223344")

            if not self.uart.is_connected():
                raise ValueError("Console Device not connected")

            r = self.uart.send_packet(id=None, packetType=OW_CMD, command=OW_CMD_HWID)
            self.uart.clear_buffer()
            # r.print_packet()
            if r.data_len == 16:
                return r.data.hex()
            else:
                return None

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def get_temperature1(self) -> float:
        """
        Retrieve the temperature reading from the TX device.

        Returns:
            float: Temperature value in Celsius.

        Raises:
            ValueError: If the UART is not connected.
            Exception: If an error occurs or the received data length is invalid.
        """
        try:
            if self.uart.demo_mode:
                return 32.4

            if not self.uart.is_connected():
                logger.error("TX Device not connected")
                return 0

            # Send the GET_TEMP command
            r = self.uart.send_packet(
                id=None, packetType=OW_POWER, command=OW_POWER_GET_TEMP1
            )
            self.uart.clear_buffer()
            # r.print_packet()

            # Check if the data length matches a float (4 bytes)
            if r.data_len == 4:
                # Unpack the float value from the received data (assuming little-endian)
                temperature = struct.unpack("<f", r.data)[0]
                # Truncate the temperature to 2 decimal places
                truncated_temperature = round(temperature, 2)
                return truncated_temperature
            else:
                raise ValueError("Invalid data length received for temperature")
        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

    def get_temperature2(self) -> float:
        """
        Retrieve the temperature reading from the TX device.

        Returns:
            float: Temperature value in Celsius.

        Raises:
            ValueError: If the UART is not connected.
            Exception: If an error occurs or the received data length is invalid.
        """
        try:
            if self.uart.demo_mode:
                return 32.4

            if not self.uart.is_connected():
                logger.error("TX Device not connected")
                return 0

            # Send the GET_TEMP command
            r = self.uart.send_packet(
                id=None, packetType=OW_POWER, command=OW_POWER_GET_TEMP2
            )
            self.uart.clear_buffer()
            # r.print_packet()

            # Check if the data length matches a float (4 bytes)
            if r.data_len == 4:
                # Unpack the float value from the received data (assuming little-endian)
                temperature = struct.unpack("<f", r.data)[0]
                # Truncate the temperature to 2 decimal places
                truncated_temperature = round(temperature, 2)
                return truncated_temperature
            else:
                raise ValueError("Invalid data length received for temperature")
        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

    def turn_12v_off(self):
        try:
            if self.uart.demo_mode:
                return True

            if not self.uart.is_connected():
                raise ValueError("Console not connected")

            logger.info("Turning off 12V.")

            r = self.uart.send_packet(
                id=None, packetType=OW_POWER, command=OW_POWER_12V_OFF
            )
            self.uart.clear_buffer()
            # r.print_packet()

            if r.packet_type == OW_ERROR:
                logger.error("Error turning off 12V")
                return False
            else:
                self.is_12v_on = False
                logger.info("12V turned off successfully.")
                return True

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def turn_12v_on(self):
        try:
            if self.uart.demo_mode:
                return True

            if not self.uart.is_connected():
                raise ValueError("Console not connected")

            logger.info("Turning on 12V.")

            r = self.uart.send_packet(
                id=None, packetType=OW_POWER, command=OW_POWER_12V_ON
            )
            self.uart.clear_buffer()
            # r.print_packet()

            if r.packet_type == OW_ERROR:
                logger.error("Error turning on 12V")
                return False
            else:
                self.is_12v_on = True
                logger.info("12V turned on successfully.")
                return True

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def get_12v_status(self):
        try:
            if self.uart.demo_mode:
                return True

            if not self.uart.is_connected():
                raise ValueError("Console not connected")

            logger.info("Get 12V voltage status.")

            r = self.uart.send_packet(
                id=None, packetType=OW_POWER, command=OW_POWER_GET_12VON
            )
            self.uart.clear_buffer()
            # r.print_packet()

            if r.packet_type == OW_ERROR:
                logger.error("Error retrieving 12V status")
                return False
            else:
                if r.reserved == 1:
                    self.is_12v_on = True
                else:
                    self.is_12v_on = False

                return self.is_12v_on

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def turn_hv_on(self):
        """
        Turn on the high voltage.
        """
        try:
            if self.uart.demo_mode:
                return True

            if not self.uart.is_connected():
                raise ValueError("Console not connected")

            logger.info("Turning on high voltage.")

            r = self.uart.send_packet(
                id=None, packetType=OW_POWER, command=OW_POWER_HV_ON, timeout=30
            )
            self.uart.clear_buffer()
            # r.print_packet()

            if r.packet_type == OW_ERROR:
                logger.error("Error turning on HV Supply")
                return False
            else:
                self.is_hv_on = True
                logger.info("HV Supply turned on successfully.")
                return True

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def turn_hv_off(self):
        """
        Turn off the high voltage.
        """
        try:
            if self.uart.demo_mode:
                return True

            if not self.uart.is_connected():
                raise ValueError("Console not connected")

            logger.info("Turning off high voltage.")

            r = self.uart.send_packet(
                id=None, packetType=OW_POWER, command=OW_POWER_HV_OFF
            )
            self.uart.clear_buffer()
            # r.print_packet()

            if r.packet_type == OW_ERROR:
                logger.error("Error turning off HV Supply")
                return False
            else:
                self.is_hv_on = False
                logger.info("HV Supply turned off successfully.")
                return True

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def get_hv_status(self):
        try:
            if self.uart.demo_mode:
                return True

            if not self.uart.is_connected():
                raise ValueError("Console not connected")

            logger.info("Get high voltage status.")

            r = self.uart.send_packet(
                id=None, packetType=OW_POWER, command=OW_POWER_GET_HVON
            )
            self.uart.clear_buffer()
            # r.print_packet()

            if r.packet_type == OW_ERROR:
                logger.error("Error retrievinging HV Status")
                return False
            else:
                if r.reserved == 1:
                    self.is_hv_on = True
                else:
                    self.is_hv_on = False

                return self.is_hv_on

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def set_voltage(self, voltage: float) -> bool:
        """
        Set the output voltage.

        Args:
            voltage (float): The desired output voltage.

        Raises:
            ValueError: If the controller is not connected or voltage exceeds supply voltage.
        """
        if self.uart.demo_mode:
            return True

        if not self.uart.is_connected():
            raise ValueError("High voltage controller not connected")

        # Validate and process the DAC input
        if voltage is None:
            voltage = 0
        elif not (5.0 <= voltage <= 100.0):
            raise ValueError(
                "Voltage input must be within the valid range 5 to 100 Volts)."
            )

        try:
            dac_input = int(((voltage) / 162) * 4095)
            # logger.info("Setting DAC Value %d.", dac_input)
            # Pack the 12-bit DAC input into two bytes
            data = bytes(
                [
                    (dac_input >> 8) & 0xFF,  # High byte (most significant bits)
                    dac_input & 0xFF,  # Low byte (least significant bits)
                ]
            )

            r = self.uart.send_packet(
                id=None, packetType=OW_POWER, command=OW_POWER_SET_HV, data=data
            )
            self.uart.clear_buffer()
            # r.print_packet()

            if r.packet_type == OW_ERROR:
                logger.error("Error setting HV")
                return False
            else:
                self.supply_voltage = voltage
                logger.info("Output voltage set to %.2fV successfully.", voltage)
                return True

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def set_dacs(self, hvp: int, hvm: int, hrp: int, hrm: int) -> bool:
        """
        Set the output voltage.

        Args:
            voltage (float): The desired output voltage.

        Raises:
            ValueError: If the controller is not connected or voltage exceeds supply voltage.
        """
        if self.uart.demo_mode:
            return True

        if not self.uart.is_connected():
            raise ValueError("High voltage controller not connected")

        # Validate and process the DAC input
        if hvp is None:
            hvp = 0
        elif not (0 <= hvp <= 4095):
            raise ValueError("Dac hvp input range is 0 to 4095.")

        if hvm is None:
            hvm = 0
        elif not (0 <= hvm <= 4095):
            raise ValueError("Dac hvm input range is 0 to 4095.")

        if hrp is None:
            hrp = 0
        elif not (0 <= hrp <= 4095):
            raise ValueError("Dac hrp input range is 0 to 4095.")

        if hrm is None:
            hrm = 0
        elif not (0 <= hrm <= 4095):
            raise ValueError("Dac hrm input range is 0 to 4095.")

        try:
            # logger.info("Setting DAC Value %d.", dac_input)
            # Pack the 12-bit DAC input into two bytes
            data = bytes(
                [
                    (hvp >> 8) & 0xFF,  # High byte (most significant bits)
                    hvp & 0xFF,  # Low byte (least significant bits)
                    (hrp >> 8) & 0xFF,  # High byte (most significant bits)
                    hrp & 0xFF,  # Low byte (least significant bits)
                    (hvm >> 8) & 0xFF,  # High byte (most significant bits)
                    hvm & 0xFF,  # Low byte (least significant bits)
                    (hrm >> 8) & 0xFF,  # High byte (most significant bits)
                    hrm & 0xFF,  # Low byte (least significant bits)
                ]
            )

            r = self.uart.send_packet(
                id=None, packetType=OW_POWER, command=OW_POWER_SET_DACS, data=data
            )
            self.uart.clear_buffer()
            # r.print_packet()

            if r.packet_type == OW_ERROR:
                logger.error("Error setting DACS")
                return False
            else:
                return True

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def get_voltage(self) -> float:
        """
        Get the current output voltage setting.

        Returns:
            float: The current output voltage.

        Raises:
            ValueError: If the controller is not connected.
        """
        if not self.uart.is_connected():
            raise ValueError("High voltage controller not connected")

        try:
            if self.uart.demo_mode:
                return 18.4

            logger.info("Getting current output voltage.")

            r = self.uart.send_packet(
                id=None, packetType=OW_POWER, command=OW_POWER_GET_HV
            )
            self.uart.clear_buffer()
            # r.print_packet()

            if r.packet_type == OW_ERROR:
                logger.error("Error Getting HV Voltage reading")
                return 0.0
            elif r.data_len == 4:
                # Unpack the float value from the received data (assuming little-endian)
                voltage = struct.unpack("<f", r.data)[0]
                # Truncate the temperature to 2 decimal places
                truncated_voltage = round(voltage, 2)
                return truncated_voltage
            else:
                logger.error("Error getting output voltage from device")
                return 0.0

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def set_fan_speed(self, fan_id: int = 0, fan_speed: int = 50) -> int:
        """
        Get the current output fan percentage.

        Args:
            fan_id (int): The desired fan to set (default is 0). bottom fans (0), and top fans (1).
            fan_speed (int): The desired fan speed (default is 50).

        Returns:
            int: The current output fan percentage.

        Raises:
            ValueError: If the controller is not connected.
        """
        if not self.uart.is_connected():
            raise ValueError("High voltage controller not connected")

        if fan_id not in [0, 1]:
            raise ValueError("Invalid fan ID. Must be 0 or 1")

        if fan_speed not in range(101):
            raise ValueError("Invalid fan speed. Must be 0 to 100")

        try:
            if self.uart.demo_mode:
                return 40

            logger.info("Getting current output voltage.")

            data = bytes(
                [
                    fan_speed & 0xFF,  # Low byte (least significant bits)
                ]
            )

            r = self.uart.send_packet(
                id=None,
                addr=fan_id,
                packetType=OW_POWER,
                command=OW_POWER_SET_FAN,
                data=data,
            )

            self.uart.clear_buffer()
            # r.print_packet()

            if r.packet_type == OW_ERROR:
                logger.error("Error setting Fan Speed")
                return -1

            logger.info(f"Set fan speed to {fan_speed}")
            return fan_speed

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def get_fan_speed(self, fan_id: int = 0) -> int:
        """
        Get the current output fan percentage.

        Args:
            fan_id (int): The desired fan to read (default is 0). bottom fans (0), and top fans (1).

        Returns:
            int: The current output fan percentage.

        Raises:
            ValueError: If the controller is not connected.
        """
        if not self.uart.is_connected():
            raise ValueError("High voltage controller not connected")

        if fan_id not in [0, 1]:
            raise ValueError("Invalid fan ID. Must be 0 or 1")

        try:
            if self.uart.demo_mode:
                return 40.0

            logger.info("Getting current output voltage.")

            r = self.uart.send_packet(
                id=None, addr=fan_id, packetType=OW_POWER, command=OW_POWER_GET_FAN
            )

            self.uart.clear_buffer()
            # r.print_packet()

            if r.packet_type == OW_ERROR:
                logger.error("Error setting HV")
                return 0.0

            elif r.data_len == 1:
                fan_value = r.data[0]
                logger.info(f"Output fan speed is {fan_value}")
                return fan_value
            else:
                logger.error("Error getting output voltage from device")
                return -1

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def set_rgb_led(self, rgb_state: int) -> int:
        """
        Set the RGB LED state.

        Args:
            rgb_state (int): The desired RGB state (0 = OFF, 1 = RED, 2 = BLUE, 3 = GREEN).

        Returns:
            int: The current RGB state after setting.

        Raises:
            ValueError: If the controller is not connected or the RGB state is invalid.
        """
        if not self.uart.is_connected():
            raise ValueError("High voltage controller not connected")

        if rgb_state not in [0, 1, 2, 3]:
            raise ValueError(
                "Invalid RGB state. Must be 0 (OFF), 1 (RED), 2 (BLUE), or 3 (GREEN)"
            )

        try:
            if self.uart.demo_mode:
                return rgb_state

            logger.info("Setting RGB LED state.")

            # Send the RGB state as the reserved byte in the packet
            r = self.uart.send_packet(
                id=None,
                reserved=rgb_state & 0xFF,  # Send the RGB state as a single byte
                packetType=OW_POWER,
                command=OW_POWER_SET_RGB,
            )

            self.uart.clear_buffer()

            if r.packet_type == OW_ERROR:
                logger.error("Error setting RGB LED state")
                return -1

            logger.info(f"Set RGB LED state to {rgb_state}")
            return rgb_state

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def get_rgb_led(self) -> int:
        """
        Get the current RGB LED state.

        Returns:
            int: The current RGB state (0 = OFF, 1 = RED, 2 = BLUE, 3 = GREEN).

        Raises:
            ValueError: If the controller is not connected.
        """
        if not self.uart.is_connected():
            raise ValueError("High voltage controller not connected")

        try:
            if self.uart.demo_mode:
                return 1  # Default to RED in demo mode

            logger.info("Getting current RGB LED state.")

            r = self.uart.send_packet(
                id=None, packetType=OW_POWER, command=OW_POWER_GET_RGB
            )

            self.uart.clear_buffer()

            if r.packet_type == OW_ERROR:
                logger.error("Error getting RGB LED state")
                return -1

            rgb_state = r.reserved
            logger.info(f"Current RGB LED state is {rgb_state}")
            return rgb_state

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def soft_reset(self) -> bool:
        """
        Perform a soft reset on the Console device.

        Returns:
            bool: True if the reset was successful, False otherwise.

        Raises:
            ValueError: If the UART is not connected.
            Exception: If an error occurs while resetting the device.
        """
        try:
            if self.uart.demo_mode:
                return True

            if not self.uart.is_connected():
                raise ValueError("Console Device  not connected")

            r = self.uart.send_packet(
                id=None, packetType=OW_CMD, command=OW_CMD_RESET
            )
            self.uart.clear_buffer()
            # r.print_packet()
            if r.packet_type == OW_ERROR:
                logger.error("Error resetting device")
                return False
            else:
                return True

        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle

    def enter_dfu(self) -> bool:
        """
        Perform a soft reset to enter DFU mode on TX device.

        Returns:
            bool: True if the reset was successful, False otherwise.

        Raises:
            ValueError: If the UART is not connected.
            Exception: If an error occurs while resetting the device.
        """
        try:
            if self.uart.demo_mode:
                return True

            if not self.uart.is_connected():
                raise ValueError("TX Device not connected")

            r = self.uart.send_packet(id=None, packetType=OW_CMD, command=OW_CMD_DFU)
            self.uart.clear_buffer()
            # r.print_packet()
            if r.packet_type == OW_ERROR:
                logger.error("Error setting DFU mode for device")
                return False
            else:
                return True
        except ValueError as v:
            logger.error("ValueError: %s", v)
            raise  # Re-raise the exception for the caller to handle

        except Exception as e:
            logger.error("Unexpected error during process: %s", e)
            raise  # Re-raise the exception for the caller to handle



from __future__ import annotations

import random

from openlifu.io.LIFUInterface import LIFUInterface


def run_test(interface, iterations):
    """
    Run the LIFU test loop with random trigger settings.

    Args:
        interface (LIFUInterface): The LIFUInterface instance.
        iterations (int): Number of iterations to run.
    """
    for i in range(iterations):
        print(f"Starting Test Iteration {i + 1}/{iterations}...")

        try:
            tx_connected, hv_connected = interface.is_device_connected()
            if not tx_connected: # or not hv_connected:
                raise ConnectionError(f"LIFU Device NOT Fully Connected. TX: {tx_connected}, HV: {hv_connected}")

            print("Ping the device")
            interface.txdevice.ping()

            print("Toggle LED")
            interface.txdevice.toggle_led()

            print("Get Version")
            version = interface.txdevice.get_version()
            print(f"Version: {version}")

            print("Echo Data")
            echo, length = interface.txdevice.echo(echo_data=b'Hello LIFU!')
            if length > 0:
                print(f"Echo: {echo.decode('utf-8')}")
            else:
                raise ValueError("Echo failed.")

            print("Get HW ID")
            hw_id = interface.txdevice.get_hardware_id()
            print(f"HWID: {hw_id}")

            print("Get Temperature")
            temperature = interface.txdevice.get_temperature()
            print(f"Temperature: {temperature} °C")

            print("Get Trigger")
            trigger_setting = interface.txdevice.get_trigger()
            if trigger_setting:
                print(f"Trigger Setting: {trigger_setting}")
            else:
                raise ValueError("Failed to get trigger setting.")

            print("Set Trigger with Random Parameters")
            # Generate random trigger frequency and pulse width
            trigger_frequency = random.randint(5, 25)  # Random frequency between 5 and 25 Hz
            trigger_pulse_width = random.randint(10, 30) * 1000  # Random pulse width between 10 and 30 ms (convert to µs)

            json_trigger_data = {
                "TriggerFrequencyHz": trigger_frequency,
                "TriggerPulseCount": 0,
                "TriggerPulseWidthUsec": trigger_pulse_width,
                "TriggerPulseTrainInterval": 0,
                "TriggerPulseTrainCount": 0,
                "TriggerMode": 1,
                "ProfileIndex": 0,
                "ProfileIncrement": 0
            }
            trigger_setting = interface.txdevice.set_trigger_json(data=json_trigger_data)

            trigger_setting = interface.txdevice.get_trigger_json()
            if trigger_setting:
                print(f"Trigger Setting Applied: Frequency = {trigger_frequency} Hz, Pulse Width = {trigger_pulse_width // 1000} ms")
                if trigger_setting["TriggerFrequencyHz"] != trigger_frequency or trigger_setting["TriggerPulseWidthUsec"] != trigger_pulse_width:
                    raise ValueError("Failed to set trigger setting.")
            else:
                raise ValueError("Failed to set trigger setting.")

            print(f"Iteration {i + 1} passed.\n")

        except Exception as e:
            print(f"Test failed on iteration {i + 1}: {e}")
            break

if __name__ == "__main__":
    print("Starting LIFU Test Script...")
    interface = LIFUInterface()

    # Number of iterations to run
    test_iterations = 1000  # Change this to the desired number of iterations

    run_test(interface, test_iterations)



from __future__ import annotations


class LIFUSignal:
    def __init__(self):
        # Initialize a list to store connected slots (callback functions)
        self._slots = []

    def connect(self, slot):
        """
        Connect a slot (callback function) to the signal.

        Args:
            slot (callable): A callable to be invoked when the signal is emitted.
        """
        if callable(slot) and slot not in self._slots:
            self._slots.append(slot)

    def disconnect(self, slot):
        """
        Disconnect a slot (callback function) from the signal.

        Args:
            slot (callable): The callable to disconnect.
        """
        if slot in self._slots:
            self._slots.remove(slot)

    def emit(self, *args, **kwargs):
        """
        Emit the signal, invoking all connected slots.

        Args:
            *args: Positional arguments to pass to the connected slots.
            **kwargs: Keyword arguments to pass to the connected slots.
        """
        for slot in self._slots:
            slot(*args, **kwargs)



from __future__ import annotations

import sys
import time

from openlifu.io.LIFUInterface import LIFUInterface

# set PYTHONPATH=%cd%\src;%PYTHONPATH%
# python notebooks/test_pwr.py

print("Starting LIFU Test Script...")

interface = LIFUInterface()
tx_connected, hv_connected = interface.is_device_connected()

if not tx_connected:
    print("TX device not connected. Attempting to turn on 12V...")
    interface.hvcontroller.turn_12v_on()

    # Give time for the TX device to power up and enumerate over USB
    time.sleep(2)

    # Cleanup and recreate interface to reinitialize USB devices
    interface.stop_monitoring()
    del interface
    time.sleep(1)  # Short delay before recreating

    print("Reinitializing LIFU interface after powering 12V...")
    interface = LIFUInterface()

    # Re-check connection
    tx_connected, hv_connected = interface.is_device_connected()

if tx_connected and hv_connected:
    print("✅ LIFU Device fully connected.")
else:
    print("❌ LIFU Device NOT fully connected.")
    print(f"  TX Connected: {tx_connected}")
    print(f"  HV Connected: {hv_connected}")
    sys.exit(1)

print("Ping the device")
interface.hvcontroller.ping()

print("Starting DAC value increments...")
hvp_value = 2400
hrp_value = 2095
hrm_value = 2095
hvm_value = 2400

print(f"Setting DACs: hvp={hvp_value}, hrp={hrp_value}, hvm={hvm_value}, hrm={hrm_value}")
if interface.hvcontroller.set_dacs(hvp=hvp_value, hrp=hrp_value, hvm=hvm_value, hrm=hrm_value):
    print(f"DACs successfully set to hvp={hvp_value}, hrp={hrp_value}, hvm={hvm_value}, hrm={hrm_value}")
else:
    print("Failed to set DACs.")

print("Test HV Supply...")
print("HV OFF. Press enter to TURN ON:")
input()  # Wait for user input
if interface.hvcontroller.turn_hv_on():
    print("HV ON. Press enter to TURN OFF:")
    input()  # Wait for user input
    if interface.hvcontroller.turn_hv_off():
        print("HV OFF.")
    else:
        print("Failed to turn off HV.")
else:
    print("Failed to turn on HV.")



from __future__ import annotations

import copy
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Annotated, Any

import numpy as np
import xarray as xa

from openlifu.seg.material import MATERIALS, PARAM_INFO, Material
from openlifu.util.annotations import OpenLIFUFieldData


@dataclass
class SegmentationMethod(ABC):
    materials: Annotated[dict[str, Material], OpenLIFUFieldData("Segmentation materials", "Dictionary mapping of label names to material definitions used during segmentation")] = field(default_factory=lambda: MATERIALS.copy())
    """Dictionary mapping of label names to material definitions used during segmentation"""

    ref_material: Annotated[str, OpenLIFUFieldData("Reference material", "Reference material ID to use")] = "water"
    """Reference material ID to use"""

    def __post_init__(self):
        if self.materials is None:
            self.materials = MATERIALS.copy()
        if self.ref_material not in self.materials:
            raise ValueError(f"Reference material {self.ref_material} not found.")

    @abstractmethod
    def _segment(self, volume: xa.DataArray) -> xa.DataArray:
        pass

    def to_dict(self) -> dict[str, Any]:
        d = self.__dict__.copy()
        d['materials'] = { k: v.to_dict() for k, v in self.materials.items() }
        d['class'] = self.__class__.__name__
        return d

    @staticmethod
    def from_dict(d: dict) -> SegmentationMethod:
        from openlifu.seg import seg_methods
        if not isinstance(d, dict):  # previous implementations might pass str
            raise TypeError(f"Expected dict for from_dict, got {type(d).__name__}")

        d = copy.deepcopy(d)
        short_classname = d.pop("class")

        # Recursively construct Material instances
        materials_dict = d.get("materials")
        if materials_dict is not None:
            d["materials"] = {
                k: v if isinstance(v, Material) else Material.from_dict(v)
                for k, v in materials_dict.items()
            }

        # Ignore ref_material if class is `UniformWater` or `UniformTissue`
        if short_classname in ["UniformWater", "UniformTissue"]:
            d.pop("ref_material")
        class_constructor = getattr(seg_methods, short_classname)
        return class_constructor(**d)

    def _material_indices(self, materials: dict | None = None):
        materials = self.materials if materials is None else materials
        return {material_id: i for i, material_id in enumerate(materials.keys())}

    def _map_params(self, seg: xa.DataArray, materials: dict | None = None):
        materials = self.materials if materials is None else materials
        material_dict = self._material_indices(materials=materials)
        params = xa.Dataset()
        ref_mat = materials[self.ref_material]
        for param_id in PARAM_INFO:
            info = Material.param_info(param_id)
            param = xa.DataArray(np.zeros(seg.shape), coords=seg.coords, attrs={"units": info["units"], "long_name": info["name"], "ref_value": ref_mat.get_param(param_id)})
            for material_id, material in materials.items():
                midx = material_dict[material_id]
                param.data[seg.data == midx] = getattr(material, param_id)
            params[param_id] = param
        params.attrs['ref_material'] = ref_mat
        return params

    def seg_params(self, volume: xa.DataArray, materials: dict | None = None):
        materials = self.materials if materials is None else materials
        seg = self._segment(volume)
        params = self._map_params(seg, materials=materials)
        return params

    def ref_params(self, coords: xa.Coordinates):
        seg = self._ref_segment(coords)
        params = self._map_params(seg)
        return params

    def _ref_segment(self, coords: xa.Coordinates):
        material_dict = self._material_indices()
        m_idx = material_dict[self.ref_material]
        sz = list(coords.sizes.values())
        seg = xa.DataArray(np.full(sz, m_idx, dtype=int), coords=coords)
        return seg



from __future__ import annotations

from .focal_pattern import FocalPattern
from .single import SinglePoint
from .wheel import Wheel

__all__ = [
    "FocalPattern",
    "SinglePoint",
    "Wheel",
]
